// Description: 
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//

#include "main.h"
#include "dlgOptions.h"
#include "FreeOTFEDebug.h"
#include "FreeOTFElib.h"
#include "FreeOTFEGUIlib.h"
#include "FreeOTFE4PDAlib.h"
#include "FreeOTFE4PDAGUIlib.h"
#include "DriverInterface.h"
#include "DriverInterfaceTwo.h"
#include "DriverInterfaceCommon.h"

#include "resource.h"

#include "SDUGeneral.h"

#include <stdio.h> // Required for _snwprintf_s 
#include <winuser.h>  // Required for IDOK, etc
#include <Commdlg.h>  // Required from OPENFILENAME
#include <aygshell.h>  // Required for SH... functions/definitions/etc
#pragma comment(lib, "aygshell") // Link in aygshell.lib

// File association "Open" command
#define  FILE_ASSOC_OPEN   TEXT("\"%s\" /mount /volume \"%%1\" /freeotfe")

// Local to this file...
HWND G_dlgOptions_MenuBar = NULL;
BOOL G_dlgOptions_FileExtnAssociated = FALSE;


// =========================================================================
// Forward declarations...
BOOL CALLBACK dlgOptions_Proc(
                           HWND hDlg, 
                           UINT message, 
                           WPARAM wParam, 
                           LPARAM lParam
                          );

BOOL _dlgOptions_CheckFileAssociationPresent();


// =========================================================================
void DisplayDlgOptions(HWND hWnd)
{
    DEBUGOUTGUI(DEBUGLEV_ENTER, (TEXT("_DisplaydlgOptions\n")));
    DEBUGOUTGUI(DEBUGLEV_INFO, (TEXT("Displaying options dialog\n")));

    DialogBox(
              G_hInstance, 
              MAKEINTRESOURCE(IDD_OPTIONS), 
              hWnd, 
              (DLGPROC)dlgOptions_Proc
             );

	DEBUGOUTGUI(DEBUGLEV_EXIT, (TEXT("_DisplaydlgOptions\n")));
}                


// =========================================================================
void _dlgOptions_EnableDisableControls(HWND hDlg)
{
    BOOL specifyMntPnt;

    specifyMntPnt = (IsDlgButtonChecked(
                                        hDlg,
                                        IDC_RADIO_MNTPNT_DEFAULT
                                       ) == BST_CHECKED);
	SetControlEnabled(hDlg, IDC_STATIC_MOUNTPOINT, specifyMntPnt);
	SetControlEnabled(hDlg, IDC_EDIT_MOUNTPOINT, specifyMntPnt);
}


// =========================================================================
BOOL CALLBACK dlgOptions_HandleMsg_WM_INITDIALOG(HWND hDlg)
{
    SHINITDLGINFO shidi;

    DEBUGOUTGUI(DEBUGLEV_ENTER, (TEXT("dlgOptions_HandleMsg_WM_INITDIALOG\n")));

    shidi.dwMask = SHIDIM_FLAGS;
    shidi.dwFlags = (
                     SHIDIF_DONEBUTTON | 
                     SHIDIF_SIPDOWN | 
                     SHIDIF_SIZEDLGFULLSCREEN
                    );
    shidi.hDlg = hDlg;
    SHInitDialog(&shidi);

    G_dlgOptions_MenuBar = SetupMenu_Simple(hDlg, IDR_MENU_DONECANCEL);


    // Populate display...

    // Mountpoint radiobutton...
    if (G_Options->MntPntUseVolFilename == TRUE)
        {
        CheckDlgButton(
                       hDlg,    
                       IDC_RADIO_MNTPNT_VOL,
                       BST_CHECKED
                      );
        }
    else
        {
        CheckDlgButton(
                       hDlg, 
                       IDC_RADIO_MNTPNT_DEFAULT,
                       BST_CHECKED
                      );
        }
    SetDlgItemText(hDlg, IDC_EDIT_MOUNTPOINT, G_Options->MntPntDefault);

    SetDlgItemText(hDlg, IDC_EDIT_EXPLORER, G_Options->ExplorerExe);
    if (G_Options->ExploreOnMount == TRUE)
        {
        CheckDlgButton(
                       hDlg,    
                       IDC_CHECK_EXPLOREONMOUNT,
                       BST_CHECKED
                      );
        }

    if (G_Options->LargeIcons == TRUE)
        {
        CheckDlgButton(
                       hDlg,    
                       IDC_CHECK_LARGEICONS,
                       BST_CHECKED
                      );
        }    

    // User can never use softkeys unless running WM2005 or later
    SetControlEnabled(
                      hDlg,
                      IDC_CHECK_SOFTKEYMENUS,
                      (SDUGetOSVersion_PDA() >= SDU_OS_VER_PDA_WM2005)
                     );
    if (G_Options->SoftkeyMenus == TRUE)
        {
        CheckDlgButton(
                       hDlg,    
                       IDC_CHECK_SOFTKEYMENUS,
                       BST_CHECKED
                      );
        }    

    if (G_Options->SaveSettings == TRUE)
        {
        CheckDlgButton(
                       hDlg,    
                       IDC_CHECK_SAVESETTINGS,
                       BST_CHECKED
                      );
        }

    SetDlgItemText(hDlg, IDC_CHECK_ASSOCIATE, TEXT("Associate with \"."));
    AppendStaticText_Simple(hDlg, IDC_CHECK_ASSOCIATE, VOL_FILE_EXTN);
    AppendStaticText_Simple(hDlg, IDC_CHECK_ASSOCIATE, TEXT("\" files"));

    G_dlgOptions_FileExtnAssociated = _dlgOptions_CheckFileAssociationPresent();
    if (G_dlgOptions_FileExtnAssociated)
        {
        CheckDlgButton(
                       hDlg,    
                       IDC_CHECK_ASSOCIATE,
                       BST_CHECKED
                      );
        }

    _dlgOptions_EnableDisableControls(hDlg);

    SDUTextSetBold(hDlg, IDC_CHECK_SAVESETTINGS, TRUE);

    DEBUGOUTGUI(DEBUGLEV_EXIT, (TEXT("dlgOptions_HandleMsg_WM_INITDIALOG\n")));
    return TRUE;
}


// =========================================================================
BOOL CALLBACK dlgOptions_HandleMsg_WM_CLOSE(HWND hWnd)
{
    DEBUGOUTGUI(DEBUGLEV_ENTER, (TEXT("dlgOptions_HandleMsg_WM_CLOSE\n")));
    if (G_dlgOptions_MenuBar != NULL)
        {
        DestroyWindow(G_dlgOptions_MenuBar);
        G_dlgOptions_MenuBar = NULL;
        }
    DEBUGOUTGUI(DEBUGLEV_EXIT, (TEXT("dlgOptions_HandleMsg_WM_CLOSE\n")));
    return TRUE;
}


// =========================================================================
BOOL
dlgOptions_GetUserInput(HWND hDlg, OPTIONS* Opts)
{
	BOOL retval;
    WCHAR tmpStr[FREEOTFE_MAX_FILENAME_LENGTH];

    retval = TRUE;

    Opts->MntPntUseVolFilename = (IsDlgButtonChecked(
                                        hDlg,
                                        IDC_RADIO_MNTPNT_VOL
                                       ) == BST_CHECKED);

    SecZeroAndFreeWCHARMemory(Opts->MntPntDefault);
    GetDlgItemText( 
                   hDlg, 
                   IDC_EDIT_MOUNTPOINT,
                   tmpStr,
                   sizeof(tmpStr)
                 );
    // +1 to include terminating NULL
    Opts->MntPntDefault = calloc(
                                 (wcslen(tmpStr) + 1), 
                                 sizeof(*Opts->MntPntDefault)
                                );
    if (Opts->MntPntDefault == NULL)
        {
        DEBUGOUTGUI(DEBUGLEV_ERROR, (TEXT("Unable to allocate memory for tmp opt string\n")));
        retval = FALSE;
        }
    else
        {
        wcscpy(Opts->MntPntDefault, tmpStr);
        }

    SecZeroAndFreeWCHARMemory(Opts->ExplorerExe);
    GetDlgItemText( 
                   hDlg, 
                   IDC_EDIT_EXPLORER,
                   tmpStr,
                   sizeof(tmpStr)
                 );
    // +1 to include terminating NULL
    Opts->ExplorerExe = calloc(
                               (wcslen(tmpStr) + 1),
                               sizeof(*Opts->ExplorerExe)
                              );
    if (Opts->ExplorerExe == NULL)
        {
        DEBUGOUTGUI(DEBUGLEV_ERROR, (TEXT("Unable to allocate memory for tmp opt string\n")));
        retval = FALSE;
        }
    else
        {
        wcscpy(Opts->ExplorerExe, tmpStr);
        }
    Opts->ExploreOnMount = (IsDlgButtonChecked(
                                        hDlg,
                                        IDC_CHECK_EXPLOREONMOUNT
                                       ) == BST_CHECKED);

    Opts->LargeIcons = (IsDlgButtonChecked(
                                        hDlg,
                                        IDC_CHECK_LARGEICONS
                                       ) == BST_CHECKED);

    Opts->SoftkeyMenus = (IsDlgButtonChecked(
                                        hDlg,
                                        IDC_CHECK_SOFTKEYMENUS
                                       ) == BST_CHECKED);

    Opts->SaveSettings = (IsDlgButtonChecked(
                                        hDlg,
                                        IDC_CHECK_SAVESETTINGS
                                       ) == BST_CHECKED);

    return retval;
}


// =========================================================================
BOOL
dlgOptions_ValidateUserInput(HWND hDlg, OPTIONS* Opts)
{
	BOOL retval;

    DEBUGOUTGUI(DEBUGLEV_ENTER, (TEXT("dlgOptions_ValidateInput\n")));

	retval = TRUE;

    if (retval)
        {
        retval = ValidateMountpoint(hDlg, Opts->MntPntDefault, TRUE, FALSE);
        }

    DEBUGOUTGUI(DEBUGLEV_EXIT, (TEXT("dlgOptions_ValidateInput\n")));
	return retval;
}


// =========================================================================
// Returns: TRUE if the file extension has an association, otherwise FALSE
BOOL
_dlgOptions_CheckFileAssociationPresent()
{
    BOOL retval;

    retval = RegIsFileExtensionAssociated(NULL, VOL_FILE_TYPE);

    return retval;
}


// =========================================================================
void
_dlgOptions_FileAssociation(HWND hDlg)
{
    WCHAR tmpFilename[FREEOTFE_MAX_FILENAME_LENGTH];
    WCHAR tmpOpenCommand[FREEOTFE_MAX_FILENAME_LENGTH];
    BOOL associateSelected;
    
    // Do nothing with the file extension, unless the user's changed the
    // setting

    associateSelected = (IsDlgButtonChecked(
                           hDlg,
                           IDC_CHECK_ASSOCIATE
                          ) == BST_CHECKED);

    if (associateSelected != G_dlgOptions_FileExtnAssociated)
        {
        if (associateSelected)
            {
            if (GetModuleFileName(
                                  NULL, 
                                  tmpFilename, 
                                  sizeof(tmpFilename)
                                 ) != 0)
                {
                _snwprintf(
                           tmpOpenCommand, 
                           (sizeof(tmpOpenCommand) / sizeof(tmpOpenCommand[0])),
                           FILE_ASSOC_OPEN,
                           tmpFilename
                          );

                RegAssociateFileExtension(
                                          VOL_FILE_EXTN,
                                          VOL_FILE_TYPE,
                                          VOL_FILE_DESC,
                                          tmpOpenCommand,
                                          tmpFilename,
                                          IDI_ICON_MAIN
                                         );
                }
            }
        else
            {
            RegDisassociateFileExtension(
                                         VOL_FILE_EXTN,
                                         VOL_FILE_TYPE
                                        );
            }
        }

}


// =========================================================================
BOOL CALLBACK dlgOptions_HandleMsg_WM_COMMAND(HWND hDlg, WPARAM wParam)
{
    BOOL retval = FALSE;
	int ctrlID;

    DEBUGOUTGUI(DEBUGLEV_ENTER, (TEXT("dlgOptions_HandleMsg_WM_COMMAND\n")));
    DEBUGOUTGUI(DEBUGLEV_INFO, (TEXT("Command: %d\n"), LOWORD(wParam)));

    ctrlID = LOWORD(wParam);

    switch(ctrlID)
        {
        case IDOK:
			{
            BOOL allOK;
            OPTIONS* newOptions;

            allOK = TRUE;

            if (allOK)
                {
                newOptions = OptionsCopy(G_Options);
                allOK = (newOptions != NULL);
                }

            if (allOK)
                {
                allOK = dlgOptions_GetUserInput(hDlg, newOptions);
                }

            if (allOK)
                {
			    allOK = dlgOptions_ValidateUserInput(hDlg, newOptions);
                }

            if (allOK)
                {
                OptionsFree(G_Options);
                G_Options = newOptions;

                if (!(OptionsWrite(G_Options)))
				    {
				    MsgError(hDlg, TEXT("Options could not be saved."));
				    }

			    }
            else
                {
                OptionsFree(newOptions);
                }

            // Sort out file associations...
            _dlgOptions_FileAssociation(hDlg);

            if (allOK)
                {
                EndDialog(hDlg, TRUE);
                }

            retval = TRUE;
            break;
            }

        case IDCANCEL:
            {
            EndDialog(hDlg, TRUE);
            retval = TRUE;
            break;
            }

        case IDC_BUTTON_BROWSEEXPLORER:
            {
            OPENFILENAME ofn;
            DWORD flags;
            WCHAR filename[FREEOTFE_MAX_FILENAME_LENGTH];
			BOOL fileDlgOK;

            DEBUGOUTGUI(DEBUGLEV_INFO, (TEXT("Browse...\n")));

            memset(filename, 0, sizeof(filename));

            ofn.lStructSize       = sizeof(ofn);
            ofn.hwndOwner         = hDlg;
            ofn.hInstance         = NULL;
            ofn.lpstrCustomFilter = NULL;
            ofn.lpstrFile         = filename;
            ofn.nMaxFile          = (sizeof(filename) / sizeof(filename[0]));
            ofn.lpstrFileTitle    = NULL;
            ofn.nMaxFileTitle     = 0;
            ofn.lpstrInitialDir   = NULL;
            ofn.lpstrTitle        = NULL;
            flags = (
                     // OFN_HIDEREADONLY |
                     OFN_FILEMUSTEXIST | 
                     OFN_PATHMUSTEXIST
                    );
            ofn.Flags             = flags;
            ofn.nFileOffset       = 0;
            ofn.nFileExtension    = 0;
            ofn.lpstrDefExt       = NULL;
            ofn.lCustData         = 0;
            ofn.lpfnHook          = NULL;
            ofn.lpTemplateName    = NULL;

            ofn.lpstrFilter       = FILE_FILTER_FLT_EXECUTABLES;
            ofn.nMaxCustFilter    = FILE_FILTER_CNT_EXECUTABLES;
            ofn.nFilterIndex      = FILE_FILTER_DFLT_EXECUTABLES;

    		fileDlgOK = GetOpenFileName(&ofn);

			if (fileDlgOK)
                {
                DEBUGOUTGUI(DEBUGLEV_INFO, (TEXT("Open file dialog OK'd...\n")));
                DEBUGOUTGUI(DEBUGLEV_INFO, (TEXT("Filename: %ls\n"), filename));
                SetDlgItemText(hDlg, IDC_EDIT_EXPLORER, filename);
                }
            else
                {
                DEBUGOUTGUI(DEBUGLEV_INFO, (TEXT("Open file dialog cancelled\n")));
                }

            break;
            }

        case IDC_RADIO_MNTPNT_VOL:
        case IDC_RADIO_MNTPNT_DEFAULT:
            {
            _dlgOptions_EnableDisableControls(hDlg);
            break;
            }

        }

    DEBUGOUTGUI(DEBUGLEV_EXIT, (TEXT("dlgOptions_HandleMsg_WM_COMMAND\n")));
    return retval;
}


// =========================================================================
BOOL CALLBACK dlgOptions_Proc(
                           HWND hDlg, 
                           UINT msg, 
                           WPARAM wParam, 
                           LPARAM lParam
                          )
{
    BOOL retval = FALSE;

    switch(msg)
        {
        case WM_INITDIALOG:
            {
            retval = dlgOptions_HandleMsg_WM_INITDIALOG(hDlg);
            break;
            }

        case WM_CLOSE:
            {
            DEBUGOUTGUI(DEBUGLEV_INFO, (TEXT("MSG: WM_CLOSE\n")));
            retval = dlgOptions_HandleMsg_WM_CLOSE(hDlg);
            break;
            }

        case WM_COMMAND:
            {
            retval = dlgOptions_HandleMsg_WM_COMMAND(hDlg, wParam);
            break;
            }

        }

    return retval;
}


// =========================================================================
// =========================================================================
